/* Copyright (C) 2005-2010, Unigine Corp. All rights reserved.
 *
 * File:    tropics.cpp
 * Desc:    Tropics demo
 * Version: 1.10
 * Authors: Anna Ishina<anet@unigine.com>
 *          Valia Vaneeva <fattie@unigine.com>
 *          Alexander Zaprjagaev <frustum@unigine.com>
 *          Pavel Kurinnoy <figonet@unigine.com>
 *
 * This file is part of the Unigine engine (http://unigine.com/).
 *
 * Your use and or redistribution of this software in source and / or
 * binary form, with or without modification, is subject to: (i) your
 * ongoing acceptance of and compliance with the terms and conditions of
 * the Unigine License Agreement; and (ii) your inclusion of this notice
 * in any version of this software that you use or redistribute.
 * A copy of the Unigine License Agreement is available by contacting
 * Unigine Corp. at http://unigine.com/
 */

#include <core/unigine.h>
#include <core/scripts/benchmark.h>
#include <core/scripts/phoronix.h>
#include <core/scripts/camera.h>
#include <core/scripts/spline.h>
#include <core/scripts/tuple.h>
#include <core/scripts/time.h>
#include <scripts/common/dialogs.h>
#include <tropics_timeline.h>

//#define RELEASE
//#define BENCHMARK
//#define PHORONIX
//#define EDITOR
//#define FIXED

#define SHADOW_MAP_LOW		4
#define SHADOW_MAP_NORMAL	5

#define KEY_PAUSE			' '
#define KEY_NEXT_PATH		APP_KEY_RETURN
#define KEY_FREE_CAMERA		APP_KEY_F8
#define KEY_BENCHMARK		APP_KEY_F9

/******************************************************************************\
*
* Responsible for handling user interface.
*
\******************************************************************************/

/*
 */
namespace Interface {
	
	// init/shutdown/update.
	void init(); 
	void shutdown(); 
	void update();
	void save(Stream stream);
	void restore(Stream stream);
	
	// hides main panel
	void hidePanel();
	
	// callbacks
	void settings_clicked();
	void camera_clicked();
	void sound_clicked();
	void benchmark_clicked(); 
	void editor_clicked(); 
	void quit_clicked();
	
	WidgetVBox tropics_vbox;			// menu box
	
	WidgetButton settings_b;			// buttons
	WidgetButton camera_b;
	WidgetButton sound_b;
	WidgetButton benchmark_b;
	WidgetButton editor_b;
};

/******************************************************************************\
*
* Crab
*
\******************************************************************************/

/*
 */
class Crab {
	
	ObjectMesh island;
	ObjectMeshSkinned mesh;
	Unigine::SplinePath path;
	
	float frame = 0.0;
	float walk_frame = 0.0;
	
	float s;
	
	//
	Crab(string path_file) {
		
		// island
		island = node_editor("island1");
		
		// mesh
		mesh = node_load("tropics/nodes/crab.node");
		mesh.setNumBuffers(2);
		
		path = new Unigine::SplinePath(path_file);
		
		// alter size
		s = engine.game.getRandom(0.7,1.0);
	}
	
	~Crab() {
		
		delete mesh;
	}
	
	//
	mat4 get_transform(float frame) {
		
		mat4 transform = path.getTransform(frame);
		vec3 p0 = island.getIWorldTransform() * vec3(transform.m03,transform.m13,200.0);
		vec3 p1 = island.getIWorldTransform() * vec3(transform.m03,transform.m13,-200.0);
		
		int ret[3];
		if(island.getIntersection(p0,p1,0,ret) == 0) return transform * scale(s,s,s);
		
		return island.getWorldTransform() * setTo(ret[0],ret[0] + path.getDirection(frame),ret[1]) * rotateX(-90.0) * rotateZ(-90.0) * scale(s,s,s);
	}
	
	//
	void update() {
		
		if(mesh.isEnabled() == 0) return;
		
		frame += engine.game.getIFps() * 2.0;
		float velocity = length(path.getVelocity(frame));
		walk_frame += engine.game.getIFps() * velocity * 400.0 / s;
		
		mesh.setWorldTransform(get_transform(frame));
		
		mesh.setAnimation("tropics/animation/crab_pincer.smesh");
		mesh.setFrame(0.0);
		mesh.getBuffer(0);
		mesh.inverseBuffer(0,0);
		mesh.setFrame(frame * 10.0);
		mesh.getBuffer(1);
		mesh.mulBuffer(1,0,1);
		
		mesh.setAnimation("tropics/animation/crab_walk.smesh");
		mesh.setFrame(walk_frame);
		mesh.getBuffer(0);
		mesh.mulBuffer(0,0,1,saturate(50.0 - velocity * 1000.0));
		mesh.setBuffer(0);
	}
	
	//
	void setEnabled(int enabled) {
		mesh.setEnabled(enabled);
	}
	
	int isEnabled() {
		return mesh.isEnabled();
	}
};

/******************************************************************************\
*
* Gull
*
\******************************************************************************/

/*
 */
class Gull {
	
	ObjectMeshSkinned mesh;
	Unigine::SplinePath path;
	
	float frame = 0;
	float spline_fps = 5.0;
	float animation_fps = 25.0;
	
	float hover_timeout;
	float fly_timeout;
	int hover;
	
	//
	Gull(string path_file) {
		
		mesh = node_load("tropics/nodes/gull.node");
		mesh.setNumLayers(2);
		mesh.setLayer(0);
		mesh.setWeight(1.0);
		mesh.setAnimation("tropics/animation/gull_flight.smesh");
		mesh.setFrame(0.0);
		mesh.setLayer(1);
		mesh.setWeight(1.0);
		mesh.setAnimation("tropics/animation/gull_hover.smesh");
		mesh.setFrame(0.0);
		
		path = new Unigine::SplinePath(path_file);
		
		hover_timeout = engine.game.getRandom(3.0,6.0);
		fly_timeout = engine.game.getRandom(6.0,10.0);
	}
	
	~Gull() {
		
		delete mesh;
	}
	
	//
	void play() {
		if(mesh.getFrameTime() <= mesh.getNumFrames()) {
			int time = mesh.getFrameTime() + engine.game.getIFps() * animation_fps;
			mesh.setFrame(time);
		} else {
			mesh.setFrame(0.0);
		}
	}
	
	//
	void update() {
		
		float timer;
		
		frame += engine.game.getIFps() * spline_fps;
		mat4 transform = path.getTransform(frame);
		mesh.setWorldTransform(translate(transform * vec3_zero) * inverse(rotation(transform)) * rotateZ(-90.0));
		
		if(hover) {
			mesh.setLayer(1);
			if(timer > hover_timeout && mesh.getFrameTime() >= mesh.getNumFrames()) {
				hover = 0;
				timer = 0.0;
				fly_timeout = engine.game.getRandom(5.0,10.0);
			} else {
				play();
				timer += engine.game.getIFps();
			}
		} else {
			mesh.setLayer(0);
			if(timer > fly_timeout && mesh.getFrameTime() >= mesh.getNumFrames()) {
				hover = 1;
				timer = 0.0;
				hover_timeout = engine.game.getRandom(5.0,10.0);
			} else {
				play();
				timer += engine.game.getIFps();
			}
		}
	}
};

/******************************************************************************\
*
* Lamp
*
\******************************************************************************/

/*
 */
class Lamp {
	
	Node node;
	mat4 transform;
	float seed;
	
	/*
	 */
	Lamp(Node node) {
		
		this.node = node;
		this.transform = node.getWorldTransform();
		
		seed = engine.game.getRandom(-10000.0,10000.0);
	}
	
	~Lamp() {
		
		node = NULL;
	}
	
	/*
	 */
	float get_angle(float amplitude,float scale) {
		return engine.game.getNoise(engine.game.getTime() * scale + seed,1000.0,1.0) * amplitude;
	}
	
	void update() {
		if(node.isEnabled() == 0) return;
		node.setWorldTransform(transform * rotateX(get_angle(10.0,1.0)) * rotateY(get_angle(10.0,1.3)) * rotateZ(get_angle(20.0,0.2)));
	}
	
	/*
	 */
	void setEnabled(int enabled) {
		node.setEnabled(enabled);
	}
	
	int isEnabled() {
		return node.isEnabled();
	}
};

/* Responsible for environment changes.
 */
class Environment {
	
	Environment();
	~Environment();
	
	void setPause(int pause);
	int getPause();
	
	void setPath(int p);
	int getPath();
	
	void nextPath(int manual = 0);
	
	void setFree(int f);
	int getFree();
	
	void setSound(int s);
	int getSound();
	
	void update();
	
	void set_environment();
	void set_sun();
	void set_sound();
	void set_water_multiplier(float multiplier);
	float get_water_multiplier();
	void set_water_refraction_color(vec4 color);
	vec4 get_water_refraction_color();
	
	enum {
		NUM_PATH = 10,
	};
	
	enum {									// how path has changed
		NONE = 0,
		AUTO,
		MANUAL,
	};
	
	int pause;								// pause toggle
	int free;								// free camera toggle
	float frame;							// current frame
	float clamp = 20.0;						// clamp frames
	float fade = 60.0;						// fade frames
	float fps = 30.0;						// path fps
	
	float sound_pos;						// sound position for free camera
	float sun_pos;							// sun position for free camera
	Light sun;								// sun node
	
	PlayerDummy dummy;						// player for guided camera
	PlayerSpectator spectator;				// player for free camera
	
	int path;								// current path
	Unigine::CameraPath paths[NUM_PATH];	// environment paths
	Unigine::Triple path_data[];			// timing and events
	int path_changed;						// path is changed
	
	AmbientSource sound;					// sound
	
	Crab crabs[0];							// life forms
	Gull gulls[0];
	
	Lamp lamps_on[];						// lamps
	Lamp lamps_off[];
	int start_lights;
	
	Node huts[0];							// huts to hide
};

/******************************************************************************\
*
* Responsible for other modules management & common demo functionality.
*
\******************************************************************************/

/*
 */
namespace Demo {
	
	// init/shutdown
	void init();
	void shutdown();
	
	// update
	int update();

	// camera mode
	void setFree(int f);
	int getFree();
	
	// sound mode
	void setSound(int s);
	int getSound();
	
	// save/restore
	void save(Stream stream);
	void restore(Stream stream);
	
	// run/stop/get benchmark
	void runBenchmark();
	void stopBenchmark();
	int getBenchmark();
	
	void runPhoronix();
	void getPhoronix();
	
	Environment environment;		// environment
	
	Unigine::Benchmark benchmark;	// benchmark
	Unigine::Phoronix phoronix;		// phoronix benchmark
};

/******************************************************************************\
*
* Interface implementation
*
\******************************************************************************/

/*
 */
void Interface::init() {
	
	Gui gui = engine.getGui();
	
	// user interface
	engine.gui.addUserInterface(gui,"demos/tropics/tropics.ui");
	
	// editor button
	#ifndef EDITOR
		editor_b.setHidden(1);
	#endif
		
	// panel
	engine.gui.addChild(tropics_vbox,GUI_ALIGN_OVERLAP);
	
	// initialize all interface windows.
	#ifdef RELEASE
		LightKeyline::init(gui);
	#else
		KeylineWindow::init(gui);
		KeyEditorWindow::init(gui);	
	#endif
}

/* Shut down.
 */
void Interface::shutdown() {
	#ifdef RELEASE
		LightKeyline::shutdown();
	#else
		KeylineWindow::shutdown();
		KeyEditorWindow::shutdown();
	#endif
}

/* Update.
 */
void Interface::update() {
	
	#ifdef RELEASE
		if(LightKeyline::isVisible()) LightKeyline::update();
	#else
		// update Keyline system
		if(KeylineWindow::isVisible()) KeylineWindow::update();  
		if(KeyEditorWindow::isVisible()) KeyEditorWindow::update();
	#endif
	
	// update setting button
	if(settings_b.isToggled() != engine.system.call("systemGetActivity")) {
		if(settings_b.isFocused() == 0) {
			settings_b.setCallbackEnabled(GUI_CLICKED,0);
			settings_b.setToggled(engine.system.call("systemGetActivity"));
			settings_b.setCallbackEnabled(GUI_CLICKED,1);
		}
	}
	
	// update camera button
	if(camera_b.isToggled() != Demo::getFree()) {
		if(camera_b.isFocused() == 0) {
			camera_b.setCallbackEnabled(GUI_CLICKED,0);
			camera_b.setToggled(Demo::getFree());
			camera_b.setCallbackEnabled(GUI_CLICKED,1);
		}
	}
	
	// update sound button
	if(sound_b.isToggled() != Demo::getSound()) {
		if(sound_b.isFocused() == 0) {
			sound_b.setCallbackEnabled(GUI_CLICKED,0);
			sound_b.setToggled(Demo::getSound());
			sound_b.setCallbackEnabled(GUI_CLICKED,1);
		}
	}
	
	// update benchmark button
	if(benchmark_b.isToggled() != Demo::getBenchmark()) {
		if(benchmark_b.isFocused() == 0) {
			benchmark_b.setCallbackEnabled(GUI_CLICKED,0);
			benchmark_b.setToggled(Demo::getBenchmark());
			benchmark_b.setCallbackEnabled(GUI_CLICKED,1);
		}
	}
	
	// update editor button
	if(editor_b != NULL && editor_b.isToggled() != engine.editor.isLoaded()) {
		if(editor_b.isFocused() == 0) {
			editor_b.setCallbackEnabled(GUI_CLICKED,0);
			editor_b.setToggled(engine.editor.isLoaded());
			editor_b.setCallbackEnabled(GUI_CLICKED,1);
		}
	}
	
	// demo is running
	if(engine.editor.isLoaded() == 0) {
		tropics_vbox.setPosition(0,0);
		benchmark_b.setEnabled(1);
	}
	// editor is loaded
	else {
		tropics_vbox.setPosition(0,42);
		benchmark_b.setEnabled(0);
	}
}

/*
 */
void Interface::save(Stream stream) {
	
}

void Interface::restore(Stream stream) {
	// restore Keyline system
	#ifdef RELEASE
		LightKeyline::init(engine.getGui());
	#else
		KeylineWindow::restore();
		KeyEditorWindow::restore();
	#endif
	init();
}

/* Hide menu panel.
 */
void Interface::hidePanel() {
	tropics_vbox.setHidden(1);
}

/*****************************************************************************\
 *
 * Interface callbacks
 *
\*****************************************************************************/

/*
 */
void Interface::settings_clicked() {
	settings_b.removeFocus();
	if(settings_b.isToggled()) engine.system.call("systemShow");
	else engine.system.call("systemHide");
}

/*
 */
void Interface::camera_clicked() {
	camera_b.removeFocus();
	Demo::setFree(camera_b.isToggled());
}

/*
 */
void Interface::sound_clicked() {
	sound_b.removeFocus();
	Demo::setSound(sound_b.isToggled());
}

/*
 */
void Interface::benchmark_clicked() {
	benchmark_b.removeFocus();
	Demo::runBenchmark();
}

/*
 */
void Interface::editor_clicked() {
	editor_b.removeFocus();
	if(editor_b.isToggled()) engine.console.run("editor_load");
	else engine.console.run("editor_quit");
}

/*
 */
void Interface::quit_clicked() {
	engine.console.run("quit");
}

/******************************************************************************\
*
* Environment implementation
*
\******************************************************************************/

/*
 */
void Environment::__Environment__() {
	
	// first path
	path = 0;
	pause = 0;
	frame = 0.0;
	path_changed = AUTO;
	
	// guided player
	dummy = new PlayerDummy();
	dummy.setZNear(0.1);
	dummy.setZFar(10000.0);
	engine.game.setPlayer(dummy);
	
	// free player
	spectator = new PlayerSpectator();
	spectator.setCollisionRadius(1.0);
	spectator.setMinVelocity(8.0);
	spectator.setMaxVelocity(16.0);
	spectator.setZNear(0.1);
	spectator.setZFar(10000.0);
	
	// load paths
	forloop(int i = 0; NUM_PATH) {
		paths[i] = new Unigine::CameraPath(format("tropics/cameras/camera_%d.txt",i),1,1.0);
	}
	
	// load path data
	Xml xml = new Xml("events");
	xml.load("tropics/tropics.event");
	forloop(int i = 0; xml.getNumChilds()) {
		Xml child = xml.getChild(i);
		path_data.append(child.getIntArg("number"),new Unigine::Triple(
			child.getFloatArg("sound_time"),
			new Unigine::Pair(child.getFloatArg("demo_time"),child.getIntArg("demo_speed")),
			child.getData()));
	}
	xml.clear();
	
	// sound
	sound = new AmbientSource("tropics/tropics.ogg",1);
	sound.setLoop(1);
	
	// crabs
	int i = 1;
	File file = new File();
	while(1) {
		string path = format("tropics/splines/crab_%d.txt",i++);
		if(file.open(path,"rb") == 0) break;
		file.close();
		crabs.append(new Crab(path));
	}
	
	// gulls
	i = 1;
	while(1) {
		string path = format("tropics/splines/gull_%d.txt",i++);
		if(file.open(path,"rb") == 0) break;
		file.close();
		gulls.append(new Gull(path));
	}
	
	// lamps, sun, and huts
	forloop(int i = 0; engine.editor.getNumNodes()) {
		Node node = engine.editor.getNode(i);
		string name = node.getName();
		if(strncmp(name,"lamp_lit_",9) == 0) {
			node.setEnabled(0);
			int key = int(substr(name,9,strlen(name) - 9));
			lamps_on.append(key,new Lamp(node));
		} else if(strncmp(name,"lamp_",5) == 0) {
			int key = int(substr(name,5,strlen(name) - 5));
			node.setEnabled(1);
			lamps_off.append(key,new Lamp(node));
		} else if(name == "pier_hut_3" || name == "pier_hut_6") {
			node.setEnabled(1);
			huts.append(node);
		} else if(name == "sun") {
			node.setEnabled(1);
			sun = node_cast(node);
		}
	}
	
	// radial blur params
	Material material = engine.materials.findMaterial("post_blur_radial");
	material.setParameterSlider(0,1.0);
	
	// water params
	set_water_multiplier(0.1);
	set_water_refraction_color(vec4(0.00784,0.04706,0.07059,1.0));
	
	// dof params
	if(engine.render.getDOF() != 0) engine.console.run("render_dof 0 && render_restart");
	engine.render.setDOFFocalDistance(10.0);
	engine.render.setDOFFocalRange(5.0);
	engine.render.setDOFBlurRange(30.0);
	engine.render.setDOFBlurRadius(1.5);
	
	set_sun();
	set_environment();
}

void Environment::__~Environment__() {
	
	path_data.clear();
	crabs.clear();
	gulls.clear();
	lamps_on.clear();
	lamps_off.clear();
}

/*
 */
void Environment::setPause(int p) {
	pause = p;
}

/*
 */
int Environment::getPause() {
	return pause;
}

/*
 */
void Environment::setPath(int p) {
	path = ::clamp(p,0,NUM_PATH - 1);
	pause = 0;
	frame = 0.0;
}

int Environment::getPath() {
	return path;
}

void Environment::nextPath(int manual = 0) {
	if(++path == NUM_PATH) path = 0;
	setPath(path);
	path_changed = (manual || path == 0) ? MANUAL : AUTO;
}

/*
 */
void Environment::setFree(int f) {
	
	free = f;
	
	if(free) {
		
		// set up player
		spectator.setPosition(dummy.getPosition());
		spectator.setDirection(dummy.getDirection());
		engine.game.setPlayer(spectator);
		
		// turn off radial blur and DOF
		if(spectator.getMaterials() != "") spectator.setMaterials("");
		if(engine.render.getDOF() != 0) engine.console.run("render_dof 0 && render_restart");
		
		// freeze, remember sky settings and enable sun
		sun_pos = TropicsTimeline::getTime();
		
		TropicsTimeline::timeline.setMultiplier(2);
		TropicsTimeline::timeline.setDuration(1440);
		TropicsTimeline::setPlayMode(1);
		
		// remember sound position
		sound_pos = sound.getTime();
		
	} else {
		
		// restore player
		engine.game.setPlayer(dummy);
		
		// turn on radial blur and DOF
		if(path_data[path].third == "clouds") dummy.setMaterials("post_blur_radial");
		else if(path_data[path].third == "dof") engine.console.run("render_dof 1 && render_restart");
		
		// restore sky settings
		TropicsTimeline::setTime(sun_pos);
		TropicsTimeline::setPlayMode(1);

		// rewind sound
		sound.setTime(sound_pos);
	}
}

int Environment::getFree() {
	return free;
}

/*
 */
void Environment::setSound(int s) {
	sound.setGain((s) ? 1.0 : EPSILON);
	if(sound.isPlaying() == 0) sound.play();
}

int Environment::getSound() {
	return (sound.getGain() > EPSILON);
}

/*
 */
void Environment::update() {
	
	// demo mode
	if(engine.editor.isLoaded() == 0) {
		
		if(free == 0) {
			
			// pause sound and sun, if needed
			if(sound.isPaused() != pause) sound.pause();
			if(TropicsTimeline::getPlayMode() == pause) TropicsTimeline::setPlayMode(!pause);
			
			// update frame
			if(pause == 0) frame += engine.game.getIFps() * fps;
			
			// switch to next path
			Unigine::CameraPath p = paths[path];
			if(frame > p.getNumFrames() - clamp * 2.0) {
				nextPath();
				p = paths[path];
			}
			
			// update path
			p.setFrame(dummy,frame + clamp);
			
			if(path_changed != NONE) {
				set_environment();
				set_sun();
				if(path_changed == MANUAL) set_sound();
			} else if(path_data[path].third == "clouds") {
				// fade out radial blur
				Material material = engine.materials.findMaterial("post_blur_radial");
				if(material.getParameterSlider(0) > 1e-4) material.setParameterSlider(0,1.0 - frame * 2e-3);
			}
			
			// fade
			float f = frame / fade;
			if(f > 1.0) f = (p.getNumFrames() - clamp * 2.0 - frame) / fade;
			if(f > 1.0) f = 1.0;
			engine.render.setFadeColor(f - 1.0);
			
			// lamps
			float timer;
			if(start_lights) {
				if(path_changed) timer = 1.5;
				if(timer <= 0.0) {
					foreachkey(int key; lamps_on) {
						if(lamps_on[key].isEnabled() == 0) {
							lamps_on[key].setEnabled(1);
							if(lamps_off.check(key)) lamps_off[key].setEnabled(0);
							timer = 1.5;
							break;
						}
					}
				} else {
					timer -= engine.game.getIFps();
				}
			} else {
				timer = 0.0;
			}
		}
		
		foreach(Lamp lamp; lamps_on) {
			lamp.update();
		}
		foreach(Lamp lamp; lamps_off) {
			lamp.update();
		}
		path_changed = NONE;
	}
	else {
		
		// let level editors play their own music
		if(sound.isPlaying()) sound.pause();
		
		TropicsTimeline::setPlayMode(0);
	}
	
	foreach(Crab crab; crabs) {
		crab.update();
	}
	
	foreach(Gull gull; gulls) {
		gull.update();
	}
	
	// check shadow map resolution
	if(engine.render.getShaders() == RENDER_QUALITY_LOW && sun.getShadowSize() != SHADOW_MAP_LOW) sun.setShadowSize(SHADOW_MAP_LOW);
	else if(engine.render.getShaders() != RENDER_QUALITY_LOW && sun.getShadowSize() != SHADOW_MAP_NORMAL) sun.setShadowSize(SHADOW_MAP_NORMAL);
}

/*
 */
void Environment::set_environment() {
	
	// switch objects on/off
	foreach(Node hut; huts) {
		hut.setEnabled(path_data[path].third != "huts");
	}
	foreach(Crab crab; crabs) {
		crab.setEnabled(path_data[path].third != "crabs");
	}
	start_lights = (path_data[path].third == "lamps_gradually");
	foreach(Lamp lamp; lamps_on) {
		lamp.setEnabled(path_data[path].third == "lamps");
	}
	foreach(Lamp lamp; lamps_off) {
		lamp.setEnabled(path_data[path].third != "lamps");
	}
	
	// switch radial blur on/off
	Player player = engine.game.getPlayer();
	if(path_data[path].third == "clouds") {
		if(engine.editor.isLoaded() == 0) player.setMaterials("post_blur_radial");
		Material material = engine.materials.findMaterial("post_blur_radial");
		material.setParameterSlider(0,1.0);
	}
	else if(player.getMaterials() != "") {
		player.setMaterials("");
	}
	
	// adjust settings
	if(path_data[path].third == "clouds") engine.render.setShadowDistance(1200.0);
	else engine.render.setShadowDistance(300.0);
	
	if(path_data[path].third == "dof") engine.console.run("render_dof 1 && render_restart");
	else if(engine.render.getDOF() != 0) engine.console.run("render_dof 0 && render_restart");
	
	if(path_data[path].third == "water") set_water_multiplier(0.5);
	else set_water_multiplier(0.1);
	
	if(path_data[path].third == "lamps_gradually" || path_data[path].third == "lamps") set_water_refraction_color(vec4(0.00392,0.01961,0.03137,1.0));
	else set_water_refraction_color(vec4(0.00784,0.04706,0.07059,1.0));
}

void Environment::set_sun() {
	Unigine::Pair pair = path_data[path].second;
	TropicsTimeline::setTime(pair.first);
	TropicsTimeline::setTimeSpeed(pair.second);
}

void Environment::set_sound() {
	sound.setTime(path_data[path].first);
}

void Environment::set_water_multiplier(float multiplier) {
	Material material = engine.materials.findMaterial("tropics_water");
	int id = material.findParameter("volume_density");
	material.setParameterSlider(id,multiplier);
}

float Environment::get_water_multiplier() {
	Material material = engine.materials.findMaterial("tropics_water");
	int id = material.findParameter("volume_density");
	return material.getParameterSlider(id);
}

void Environment::set_water_refraction_color(vec4 color) {
	Material material = engine.materials.findMaterial("tropics_water");
	int id = material.findParameter("refraction_color");
	material.setParameter(id,color);
}

float Environment::get_water_refraction_color() {
	Material material = engine.materials.findMaterial("tropics_water");
	int id = material.findParameter("refraction_color");
	return material.getParameter(id);
}

/******************************************************************************\
*
* Demo implementation
*
\******************************************************************************/

/*
 */
void Demo::init() {
	
	// initialize other modules.
	TropicsTimeline::init();
	Interface::init();
	
	// create environment
	environment = new Environment();
	
	int flags = Unigine::Benchmark::RENDER_SHADERS |
		Unigine::Benchmark::RENDER_TEXTURES |
		Unigine::Benchmark::RENDER_FILTER |
		Unigine::Benchmark::RENDER_ANISOTROPY |
		Unigine::Benchmark::RENDER_OCCLUSION |
		Unigine::Benchmark::RENDER_REFLECTION |
		Unigine::Benchmark::RENDER_REFRACTION |
		Unigine::Benchmark::RENDER_VOLUMETRIC;
	
	benchmark = new Unigine::Benchmark("Tropics Demo v1.3",25.19,flags);
	phoronix = new Unigine::Phoronix();
	
	// play sound
	if(engine.config.isExist("tropics_sound")) setSound(engine.config.getBool("tropics_sound"));
	else setSound(1);
}

/*
 */
void Demo::shutdown() {
	
	engine.config.setBool("tropics_sound",getSound());
	
	Interface::shutdown();
	TropicsTimeline::shutdown();
	delete environment;
}

/*
 */
int Demo::update() {
	
	environment.update();
	
	TropicsTimeline::update();
	Interface::update();
	
	// check hot keys when only demo is running
	if(engine.editor.isLoaded() == 0 && engine.console.getActivity() == 0 && benchmark.getActivity() == 0 && phoronix.getActivity() == 0) {
		
		Gui gui = engine.getGui();
		
		if(gui.getKeyActivity(KEY_PAUSE) == 0 && engine.app.clearKeyState(KEY_PAUSE)) environment.setPause(!environment.getPause());
		if(gui.getKeyActivity(KEY_NEXT_PATH) == 0 && engine.app.clearKeyState(KEY_NEXT_PATH)) environment.nextPath(1);
		if(gui.getKeyActivity(KEY_FREE_CAMERA) == 0 && engine.app.clearKeyState(KEY_FREE_CAMERA)) environment.setFree(!environment.getFree());
		if(gui.getKeyActivity(KEY_BENCHMARK) == 0 && engine.app.clearKeyState(KEY_BENCHMARK)) {
			Interface::benchmark_b.setFocus();
			Interface::benchmark_b.setToggled(1);
		}
	}
	
	// flags
	int benchmark_path;
	int phoronix_path;
	
	// benchmark
	if(benchmark.getActivity()) {
		benchmark.update(format("%d of %d",environment.getPath() + 1,Environment::NUM_PATH));
		if(environment.getPath() == 0 && benchmark_path == Environment::NUM_PATH - 1) {
			engine.app.setMouseGrab(0);
			benchmark.done();
			#ifdef BENCHMARK
				engine.console.run("quit");
			#endif
		}
		benchmark_path = environment.getPath();
	} else {
		benchmark_path = -1;
	}	
	
	// phoronix benchmark
	if(phoronix.getActivity()) {
		phoronix.update();
		if(environment.getPath() == 0 && phoronix_path == Environment::NUM_PATH - 1) {
			phoronix.done();
			#ifdef PHORONIX
				engine.console.run("quit");
			#endif
		}
		phoronix_path = environment.getPath();
	} else {
		phoronix_path = -1;
	}
	
	// editor
	if(engine.editor.isLoaded()) {
		if(benchmark.getActivity()) benchmark.stop();
	}
	
	#ifdef _WIN32
		if(engine.app.getKeyState(APP_KEY_ALT) && engine.app.clearKeyState(APP_KEY_F4)) {
			engine.console.run("quit");
		}
	#endif
	
	if(getPhoronix() && engine.system.call("systemGetActivity") == 0) {
		engine.gui.setMouseCursor(GUI_CURSOR_NONE);
	}
}

/*
 */
void Demo::setFree(int f) {
	environment.setFree(f);
}

int Demo::getFree() {
	return environment.getFree();
}

/*
 */
void Demo::setSound(int s) {
	environment.setSound(s);
}

int Demo::getSound() {
	return environment.getSound();
}

/*
 */
void Demo::save(Stream stream) {
	Interface::save(stream);
}

void Demo::restore(Stream stream) {
	Interface::restore(stream);
}

/*
 */
void Demo::runBenchmark() {
	environment.setFree(0);
	environment.setPath(0);
	environment.path_changed = Environment::MANUAL;
	benchmark.run();
}

/*
 */
void Demo::stopBenchmark() {
	benchmark.stop();
}

/*
 */
int Demo::getBenchmark() {
	return benchmark.getActivity();
}

void Demo::runPhoronix() {
	if(engine.editor.isLoaded()) engine.console.run("editor_quit");
	environment.setFree(0);
	environment.setPath(0);
	Interface::hidePanel();
	phoronix.run();
}

int Demo::getPhoronix() {
	return phoronix.getActivity();
}

/******************************************************************************\
*
* init/shutdown/update implementation
*
\******************************************************************************/

/*
 */
int init() {
	
	engine.console.run("render_hdr 1 && render_srgb 0 && render_restart");
	
	Demo::init();
	
	#ifdef RELEASE
		log.message("Release\n");
		TropicsTimeline::setPlayMode(1);
	#endif
	
	#ifdef BENCHMARK
		Demo::runBenchmark();
	#endif
	
	#ifdef PHORONIX
		Demo::runPhoronix();
	#endif
	
	#ifdef FIXED
		engine.game.setIFps(1.0 / 30.0);
		engine.render.setHDRAdaptation(0.0);
	#endif
	
	return 1;
}

int shutdown() {
	Demo::shutdown();
	return 1;
}

int update() {
	Demo::update();
	return 1;
}

/*
 */
int save(Stream stream) {
	Demo::save(stream);
	return 1;
}

int restore(Stream stream) {
	Demo::restore(stream);
	return 1;
}
