/* Copyright (C) 2005-2010, Unigine Corp. All rights reserved.
 *
 * File:	tropics_params.h
 * Desc:	Tropics parameter classes and editors.
 * Version:	0.50
 * Authors:	Simonov Vasiliy <simonowvv@sibmail.com>
 * 
 * This file is part of the Unigine engine (http://unigine.com/).
 *
 * Your use and or redistribution of this software in source and / or
 * binary form, with or without modification, is subject to: (i) your
 * ongoing acceptance of and compliance with the terms and conditions of
 * the Unigine License Agreement; and (ii) your inclusion of this notice
 * in any version of this software that you use or redistribute.
 * A copy of the Unigine License Agreement is available by contacting
 * Unigine Corp. at http://unigine.com/
 */

#ifndef __TROPICS_PARAMS_H__
#define __TROPICS_PARAMS_H__

#include <keyline/params.h>

/*****************************************************************************\
 *
 * forwards
 *
\*****************************************************************************/

void registerTropicsParams();
KeyLine::ParamBase param_sun_clone(KeyLine::ParamBase src);
KeyLine::ParamBase param_environment_clone(KeyLine::ParamBase src);

/*************************************************************************\
*
* namespace Sky
*
\*************************************************************************/

/* namespace for sky and sun settings
 */
namespace Sky {
	
	// variables
	ObjectSky sky = NULL;
	LightWorld sun = NULL;
	Unigine::Time time = NULL; // time singleton
	int sphere = 0;
	int shadow = 0;
	int initialized = false;
	
	/* init static variables in Sky namespace
	 */
	void init() {
		if(!initialized) {
			// init time
			time = new Unigine::Time(1.0f,-8.2f,-8.2f);
			time.setDate(2009,7,1);
			// find sun and sky
			forloop(int i = 0; engine.editor.getNumNodes()) {
				Node node = engine.editor.getNode(i);
				string type = node.getTypeName();
				if(type == "ObjectSky") {
					sky = class_cast("ObjectSky",node);
				}
				else if(type == "LightWorld") {
					sun = class_cast("LightWorld",node);
				}
				// all objects found
				if(sun != NULL && sky != NULL) break;
			}
			
			// search for sky surfaces
			if(sky != NULL) {
				forloop(int i = 0; sky.getNumSurfaces()) {
					if(sky.getSurfaceName(i) == "sphere") {
						sphere = i;
					}
					else if(sky.getSurfaceName(i) == "shadow") {
						shadow = i;
					}
				}
			}
			initialized = true;
		}
	}
	
	/* returns global sky object
	 */
	ObjectSky getSky() {
		if(!initialized) init();
		return sky;
	}
	
	/* returns global sun object
	 */
	LightWorld getSun() {
		if(!initialized) init();
		return sun;
	}
	
	/* get/set time for sun
	 */
	Unigine::Time getTime() { return time; }
	void setTime(Unigine::Time t) { 
		if(time == NULL || sun == NULL) return;
		time = t; 
		sun.setWorldTransform(time.getSunTransform());
	}
	
	/* set amount of secons to time variable
	 */
	void setTimeSeconds(float sec) {
		Unigine::Time t = getTime();
		t.setSeconds(sec);
		setTime(t);
	}
	
	
	/*************************************************************************\
	*
	* sun parameters getters/setters
	*
	\*************************************************************************/
	
	//
	float getSunShadowAmbient() { return sun.getShadowAmbient(); }
	void setSunShadowAmbient(float value) { sun.setShadowAmbient(value); }
	
	//
	float getSunMultiplier() { return sun.getMultiplier(); }
	void setSunMultiplier(float value) { sun.setMultiplier(value); }
	
	//
	vec4 getSunColor() { return sun.getColor(); }
	void setSunColor(vec4 value) { sun.setColor(value); }
	
	//
	float getSunSpecularScale() { return sun.getSpecularScale(); }
	void setSunSpecularScale(float value) { sun.setSpecularScale(value); }
	
	//
	float getSunSpecularPower() { return sun.getSpecularPower(); }
	void setSunSpecularPower(float value) { sun.setSpecularPower(value); }
	
	//
	float getShadowScale() { vec4 vec = sky.getMaterialParameter("shadow_scale",shadow); return vec.x; }
	void setShadowScale(float value) { sky.setMaterialParameter("shadow_scale",vec4(value,0.0,0.0,0.0),shadow); }
	
	//
	float getClouds01Scale() { vec4 vec = sky.getMaterialParameter("clouds_01_scale",sphere); return vec.x; }
	void setClouds01Scale(float value) { sky.setMaterialParameter("clouds_01_scale",vec4(value,0.0,0.0,0.0),sphere); }
	
	//
	float getClouds23Scale() { vec4 vec = sky.getMaterialParameter("clouds_23_scale",sphere); return vec.x; }
	void setClouds23Scale(float value) { sky.setMaterialParameter("clouds_23_scale",vec4(value,0.0,0.0,0.0),sphere); }
	
	//
	float getClouds01Threshold() { vec4 vec = sky.getMaterialParameter("clouds_01_threshold",sphere); return vec.x; }
	void setClouds01Threshold(float value) { sky.setMaterialParameter("clouds_01_threshold",vec4(value,0.0,0.0,0.0),sphere); }
	
	//
	float getClouds23Threshold() { vec4 vec = sky.getMaterialParameter("clouds_23_threshold",sphere); return vec.x; }
	void setClouds23Threshold(float value) { sky.setMaterialParameter("clouds_23_threshold",vec4(value,0.0,0.0,0.0),sphere); }
}

/*****************************************************************************\
 *
 * class ParamSun
 *
\*****************************************************************************/

/* special param type for sun position. Use golbal sun and sky.
 * hold only time value to set up sun position, if needed
 */ 
class ParamSun : KeyLine::ParamBase {
	
	// data
	float time;
	
	//
	void apply() {
		Sky::setTimeSeconds(time);
	}
	
	//
	void retrieve() { }
	
	//
	void load(Xml xml) { }
	
	//
	void save(Xml xml) { }
	
	//
	void setInterpolated(ParamBase from,ParamBase to,float weight,float pos) { time = pos; }
	
	//
	ParamBase clone() { return param_sun_clone(this); }
	
	//
	string toString() { return format("ParamSun name=\"%s\" time=%f\n",getName(),time); } 
};

/*****************************************************************************\
 *
 * class ParamEnvironment
 *
\*****************************************************************************/

/* class for environment map parameters
 */
class ParamEnvironment : KeyLine::ParamBase {
	string texture_name = "";
	vec3 coefs[9];
	
	// constructor
	ParamEnvironment(string name,string get_func = "",string set_func = "") {
		setName(name);
		setGetter(get_func);
		setSetter(set_func);
	}
	
	/* get/set coefficients
	 */ 
	void getCoefs(vec3 out_coefs[]) { out_coefs.copy(coefs); }
	void setCoefs(vec3 in_coefs[]) { coefs.copy(in_coefs); }
	
	/* get/set texture name
	 */
	string getTextureName() { return texture_name; }
	void setTextureName(string n) { texture_name = n; }
	
	//
	void apply() {
		engine.render.setEnvironmentCoefficients(coefs);
	}
	
	//
	void retrieve() {
		texture_name = engine.render.getEnvironmentTextureName();
		engine.render.getEnvironmentCoefficients(coefs);
	}
	
	//
	void load(Xml xml) {
		texture_name = xml.getData();
		// retrieve coeffecients
		engine.render.setEnvironmentTextureName(texture_name);
		engine.render.getEnvironmentCoefficients(coefs);
	}
	
	//
	void save(Xml xml) {
		xml.setData(texture_name);
	} 
	
	//
	void setInterpolated(ParamBase from,ParamBase to,float weight,float pos) {
		assert(typeof(this) == typeof(from) && typeof(from) == typeof(to));
		ParamEnvironment src = from;
		ParamEnvironment dest = to;
		// get texture name from previous
		setTextureName(src.getTextureName());
		// interpolate coeffecients
		forloop(int i = 0; coefs.size()) {
			coefs[i] = lerp(src.coefs[i],dest.coefs[i],weight);
		}
	}
	
	//
	ParamBase clone() {
		return param_environment_clone(this);
	}
	
	//
	string toString() { 
		return format("ParamEnvironment \"%s\" = %s",getName(),texture_name);
	}
};

/*****************************************************************************\
 *
 * class EnvironmentEditor
 *
\*****************************************************************************/

/* class for editing ParamEnvironment
 */
class ParamEnvironmentEditor : KeyLine::ParamEditorBase {
	// data
	ParamEnvironment data = NULL;
	// controls
	WidgetHBox parent;
	WidgetLabel label;
	WidgetEditLine edit;
	WidgetIcon load_icon;
	WidgetIcon view_icon;
	WidgetIcon clear_icon;
	
	// constructor
	ParamEnvironmentEditor(Gui gui,string caption) {
		setName(caption);
		// init controls
		parent = new WidgetHBox(gui,SPACE_X,SPACE_Y);
		// init label
		label = new WidgetLabel(gui,getName() + ":");
		label.setWidth(HEADER_WIDTH);
		parent.addChild(label,GUI_ALIGN_RIGHT);
		// init edit
		edit = new WidgetEditLine(gui);
		edit.setCallback(GUI_PRESSED,"ParamEnvironmentEditor::on_editline_changed",this);
		edit.setCallback(GUI_FOCUS_OUT,"ParamEnvironmentEditor::on_editline_changed",this);
		parent.addChild(edit,GUI_ALIGN_EXPAND);
		// init load texture icon
		load_icon = new WidgetIcon(gui,"core/editor/gui/action_load.png");
		load_icon.setWidth(ICON_SIZE);
		load_icon.setHeight(ICON_SIZE);
		load_icon.setCallback(GUI_CLICKED,"ParamEnvironmentEditor::load_icon_clicked",this);
		parent.addChild(load_icon);
		// init view icon
		view_icon = new WidgetIcon(gui,"core/editor/gui/action_view.png");
		view_icon.setWidth(ICON_SIZE);
		view_icon.setHeight(ICON_SIZE);
		view_icon.setCallback(GUI_CLICKED,"ParamEnvironmentEditor::view_icon_clicked",this);
		parent.addChild(view_icon);
		// init clear icon
		clear_icon = new WidgetIcon(gui,"core/editor/gui/action_clear.png");
		clear_icon.setWidth(ICON_SIZE);
		clear_icon.setHeight(ICON_SIZE);
		clear_icon.setCallback(GUI_CLICKED,"ParamEnvironmentEditor::clear_icon_clicked",this);
		parent.addChild(clear_icon);
	}
	
	// destructor
	~ParamEnvironmentEditor() {
		delete clear_icon;
		delete view_icon;
		delete load_icon;
		delete edit;
		delete label;
		delete parent;
	}
	
	void update() {
		// update from data
		if(data != NULL && !edit.isFocused()) 
			edit.setText(data.getTextureName());
	}
	
	ParamBase getValue() { return data; }
	void setValue(ParamBase val) { data = val; }
	
	Widget getControl() { return parent; }
	
	// callbacks
	
	void on_editline_changed(ParamEnvironmentEditor sender) {
		string tex_name = sender.edit.getText();
		if(sender.data != NULL && sender.data.getTextureName() != tex_name) {
			sender.data.setTextureName(tex_name == "" ? NULL : tex_name);
			sender.on_changed_event();
		}
	}
	
	void load_icon_clicked(ParamEnvironmentEditor sender) {
		string name = sender.edit.getText();
		if(DialogFile::show("Select render environment texture",name,".dds")) {
			sender.data.setTextureName(name);
			sender.edit.setText(name);
			sender.on_changed_event();
		}
	}
	
	void view_icon_clicked(ParamEnvironmentEditor sender) {
		if(sender.edit.getText() != "") {
			DialogImage::show("Render environment texture",sender.data.getTextureName());
		}
	}
	
	void clear_icon_clicked(ParamEnvironmentEditor sender) {
		sender.data.setTextureName(NULL);
		sender.on_changed_event();
	}
};

/*****************************************************************************\
 *
 * utility functions
 *
\*****************************************************************************/

/* make a copy of ParamSun
 */
KeyLine::ParamBase param_sun_clone(KeyLine::ParamBase src) {
	ParamSun source = src;
	ParamSun res = new ParamSun();
	res.setName(source.name);
	res.time = source.time;
	return res;
}

/* clone function for ParamEnvironment class
 */
KeyLine::ParamBase param_environment_clone(KeyLine::ParamBase src) {
	ParamEnvironment source = src;
	ParamEnvironment res = new ParamEnvironment(
		source.getName(),
		source.getGetter(),
		source.getSetter());
	res.setTextureName(source.getTextureName());
	source.getCoefs(res.coefs);
	return res;
}

/*****************************************************************************\
 *
 * interface functions
 *
\*****************************************************************************/

/* ParamSun creation function
 */
KeyLine::ParamBase createParamSun() {
	return new ParamSun();
}

/* ParamEnvironment creation function
 */
KeyLine::ParamBase createParamEnvironment() {
	return new ParamEnvironment("empty param"); 
}

/* ParamEnvironmentEditor creation function
 */
KeyLine::ParamEditorBase createParamEnvironmentEditor(KeyLine::ParamDef def,Gui gui) {
	return new ParamEnvironmentEditor(gui,def.getParamName());
}

/* register parameter type
 */
void registerTropicsParams() {
	
	Sky::init();
	
	KeyLine::addParamType("sun","createParamSun");
	KeyLine::addParamType("environment","createParamEnvironment");
	
	KeyLine::addEditorType("environment","createParamEnvironmentEditor");
}

#endif // __TROPICS_PARAMS_H__
